import time
from SimulatorFramework.TileAssembly import *
from SimulatorFramework.TileType import Tile
from SimulatorFramework.TileType import TileType
from SimulatorFramework import SimUtils
import os
import ast
import random
import itertools
from test.test_deque import BIG
from PolyominoTASTileType import Polyomino, PolyominoType, DIR_OFFSETS, labels_complement


NBRS = [(0,1,0),(1,0,0),(0,-1,0),(-1,0,0),(0,0,1),(0,0,-1)]
#==============================
# N = 0
# E = 1
# S = 2
# W = 3
# U = 4
# D = 5
#=============================#
OPP_DIR = {0:2, 1:3, 2:0, 3:1, 4:5, 5:4}

EMPTY_LOC           = 0
FRONTIER_LOC        = 1
FRONTIER_LOC_OFF    = 2


class PolyominoAssembly:
    def __init__(self):
        self.NBRS = [(0,1,0),(1,0,0),(0,-1,0),(-1,0,0),(0,0,1),(0,0,-1)]
        self.DIRSTRS = ["N", "E", "S", "W", "U", "D"]
        self.OPPDIRS = [2, 3, 0, 1, 5, 4]

        self.growthInc = 20
        self.tau = 2
        self.num_polyominoes = 0
        self.locationMap = {} # map associating locations to polyomino and component
        self.frontier = {}
        #TODO: to speed things up, keep a fronterList which is a list of the frontier locations
        #      for quick random selection. Currently will have to use random.choice(list(self.frontier.values()))
        self.minX = -20
        self.maxX =  20
        self.minY = -20
        self.maxY =  20
        self.minZ = -20
        self.maxZ =  20

        self.systemDefFile = None
        self.polyominoSystem = None

    def ClearSpace(self):
        del self.locationMap
        del self.frontier
        self.num_polyominoes = 0
        self.locationMap = {}
        self.frontier = {}
        
    def GetWidth(self):
        self.setDimensions()
        return self.width

    def GetHeight(self):
        self.setDimensions()
        return self.height
    
    def GetDepth(self):
        self.setDimensions()
        return self.depth

    def GetNumTiles(self):
        return self.num_polyominoes
    
    def GetPolyominoes(self):
        polyominoes = []
        for loc in self.locationMap:
            polyomino = self.locationMap[loc]
            if not polyomino in polyominoes:
                polyominoes.append(polyomino)
        return polyominoes
    
    def setDimensions(self):
        ((minX, minY, minZ), (maxX, maxY, maxZ)) = self.getBoundingBox()
        if minX <= self.minX:
            self.minX = self.minX - self.growthInc
        if maxX >= self.maxX:
            self.maxX = self.maxX + self.growthInc
        if minY <= self.minY:
            self.minY = self.minY - self.growthInc
        if maxY >= self.maxY:
            self.maxY = self.maxY + self.growthInc
        if minZ <= self.minZ:
            self.minZ = self.minZ - self.growthInc
        if maxZ >= self.maxZ:
            self.maxZ = self.maxZ + self.growthInc
        self.width  = self.maxX - self.minX
        self.height = self.maxY - self.minY
        self.depth  = self.maxZ - self.minZ
        self.numTiles = self.GetNumTiles()

    def getBoundingBox(self):
        min_x = min_y = min_z = max_x = max_y = max_z = 0

        if len(self.locationMap) > 0:            
            bFirst = True
            for loc in self.locationMap:
                polyomino = self.locationMap[loc]
                box = polyomino.getBoundingBox()
                
                if bFirst:
                    bFirst =  False
                    min_x, min_y, min_z = box[0]
                    max_x, max_y, max_z = box[1]
                
                if box[0][0] < min_x:
                    min_x = box[0][0]
                if box[0][1] < min_y:
                    min_y = box[0][1]
                if box[0][2] < min_z:
                    min_z = box[0][2]
                if box[1][0] > max_x:
                    max_x = box[1][0]
                if box[1][1] > max_y:
                    max_y = box[1][1]
                if box[1][2] > max_z:
                    max_z = box[1][2]

        return ((min_x, min_y, min_z), (max_x, max_y, max_z))

    def OverlapsExistingPolyomino(self, polyomino):
        for block in polyomino.blocks:
            if block in self.locationMap:
                return True
        return False

    def AddPolyomino(self, polyomino):
        if self.OverlapsExistingPolyomino(polyomino):
            SimUtils.globalApp.OutputMessage(f"Error: Attempted to add polyomino which overlaps assembly at location {block}")
            return False

        # add to location maps and remove frontier locations that are now occupied
        for block in polyomino.blocks:
            self.locationMap[block] = polyomino
            if block in self.frontier:
                self.frontier.pop(block)

        # compute new frontier locations
        for block in polyomino.blocks:
            if block in polyomino.domains:  # see if it has a glue
                domains = polyomino.domains[block]
                for xo in range(-1, 2):
                    for yo in range(-1, 2):
                        for zo in range(-1, 2):
                            if abs(xo) + abs(yo) + abs(zo) != 1:
                                continue
                            nblock = (block[0] + xo, block[1] + yo, block[2] + zo)
                            for domain in domains:
                                do = (DIR_OFFSETS[domain[1]][0], DIR_OFFSETS[domain[1]][1], DIR_OFFSETS[domain[1]][2])
                                if do == (xo,yo,zo):
                                    oppdir = (-1*do[0], -1*do[1], -1*do[2])
                                    if nblock in self.frontier:
                                        self.frontier[nblock].append((oppdir, domain[0]))
                                    else:
                                        self.frontier[nblock] = [(oppdir, domain[0])]
                         
        self.num_polyominoes += 1           
        self.setDimensions()
        #print(f'locationmap: {self.locationMap}')
        #for floc in self.frontier:
        #    nbrGlues = self.GetNeighboringGlues(floc)
        #    print(f'floc {floc} has nbrGlues {nbrGlues}')
        #print(f'loc map afer add: {self.locationMap}')

    def RemovePolyomino(self, polyomino):
        oldBlocks = []
        for block in polyomino.blocks:
            self.locationMap.pop(block)
            #  If the location of a removed block has neighboring glues, add it to the frontier
            nbrGlues = self.GetNeighboringGlues(block)
            if len(nbrGlues) > 0:
                self.frontier[block] = nbrGlues
            oldBlocks.append(block)

        #  Remove neighboring locations from the frontier if they have no other incident glues
        for block in oldBlocks:
            for xo in range(-1, 2):
                for yo in range(-1, 2):
                    for zo in range(-1, 2):
                        if abs(xo) + abs(yo) + abs(zo) != 1:
                            continue
                        nblock = (block[0] + xo, block[1] + yo, block[2] + zo)
                        if nblock in self.frontier:
                            nbrGlues = self.GetNeighboringGlues(nblock)
                            if len(nbrGlues) == 0:
                                self.frontier.pop(nblock)
                            else:
                                self.frontier[nblock] = nbrGlues
        self.num_polyominoes -= 1
        self.setDimensions()
        
    def GetNeighboringGlues(self, block):
        nbrGlues = []
        for xo in range(-1, 2):
            for yo in range(-1, 2):
                for zo in range(-1, 2):
                    if abs(xo) + abs(yo) + abs(zo) != 1:
                        continue
                    nloc = (block[0] + xo, block[1] + yo, block[2] + zo)
                    if nloc in self.locationMap:
                        npol = self.locationMap[nloc]
                        if nloc in npol.domains:
                            for domain in npol.domains[nloc]:
                                do = (DIR_OFFSETS[domain[1]][0], DIR_OFFSETS[domain[1]][1], DIR_OFFSETS[domain[1]][2])
                                oppdir = (-1*do[0], -1*do[1], -1*do[2])
                                if oppdir == (xo,yo,zo):
                                    nbrGlues.append(((xo,yo,zo), domain[0]))
        return nbrGlues
    
    def GetLoc(self, coords):
        if coords in self.locationMap:
            polyomino = self.locationMap[coords]
            return polyomino
        else:
            return None

    def GetBlockBindingInFrontierLocation(self, frontier_loc, domains):
        #print(f' for frontier loc {loc} {floc}, try domains {pct.domains[dloc]}')
        binding_strength = 0
        for adj_glue in frontier_loc:
            for domain in domains:
                do = (DIR_OFFSETS[domain[1]][0], DIR_OFFSETS[domain[1]][1], DIR_OFFSETS[domain[1]][2])
                if (do == adj_glue[0]) and (labels_complement(domain[0], adj_glue[1]) == True):
                    binding_strength += 1
        return binding_strength

    def StepForward(self):
        if len(self.frontier) == 0:
            return None

        frontier_loc_list = list(self.frontier.keys())
        start_loc = random.randrange(len(frontier_loc_list))
        
        attachable_polyominoes = set()
        for i in range(len(frontier_loc_list)):
            loc = frontier_loc_list[(start_loc+i)%len(frontier_loc_list)]
            floc = self.frontier[loc]
            for nbr_domain in floc:
            
                self.temp_polyominoes = []
                num_options = 0
                for pct_name in self.polyominoSystem.PolyominoType_dict:
                    pct = self.polyominoSystem.PolyominoType_dict[pct_name]
                    
                    for dloc in pct.domains:
                        binding_strength = self.GetBlockBindingInFrontierLocation(floc, pct.domains[dloc])
                        if binding_strength > 0:
                            trans = (0-dloc[0]+loc[0], 0-dloc[1]+loc[1], 0-dloc[2]+loc[2])
                            new_polyomino = Polyomino(pct)
                            new_polyomino.translate(trans)
                            if self.OverlapsExistingPolyomino(new_polyomino) == False:
                                # polyomino can be placed here and has at least one glue match
                                binding_strength = 1
                                if self.polyominoSystem.binding_threshold > 1:
                                    binding_strength = 0
                                    for domain_coords in new_polyomino.domains:
                                        if domain_coords in self.frontier:
                                            floc2 = self.frontier[domain_coords]
                                            domains2 = new_polyomino.domains[domain_coords]
                                            binding_strength += self.GetBlockBindingInFrontierLocation(floc2, domains2)
                                if binding_strength >= self.polyominoSystem.binding_threshold:
                                    attachable_polyominoes.add(new_polyomino)
                                    
    
        if len(attachable_polyominoes) > 0:
            SimUtils.globalApp.OutputMessage(f"{len(attachable_polyominoes)} polyomino attachments possible")
            added_polyomino = random.sample(attachable_polyominoes, 1)[0]
            self.AddPolyomino(added_polyomino)
            return [added_polyomino]
        else:
            SimUtils.globalApp.OutputMessage(f"No polyomino attachments possible")
            return None

#===============================================================================
# class PolyominoTASAssembly(Assembly):
#     def __init__(self, dims):
#         Assembly.__init__(self, width=20, height=20, tileWidths=dims, tileHeights=dims)
#         self.polyomino_types = dict()
#         self.polyomino_list = []
# 
#     def ClearSpace(self):
#         Assembly.ClearSpace(self)
#         self.polyomino_list = []
#         
#     def AddToFrontierIfNecessary(self, coords):
#         return None
#    
#     def GetFittingTilesForLoc(self, coords):
#         retSet = None
#         return retSet
#     
#     def StepForward(self):
#         return None
#         
#     def AssemblyToString(self, tileList=None):
#         assemblyString = "{"
#         return assemblyString + "}"
# 
#     def LoadTileSystem(self, fileName):
#         if len(fileName) < 1:
#             return False
#         elif os.path.isabs(fileName) == False:
#             fileName = "./" + fileName
#             fileName = os.path.abspath(fileName)
# 
#         #try:
#         os.stat(fileName)
#         rFile = open(fileName, "r")
#         SimUtils.globalApp.OutputMessage("Using tile system definition file '" + fileName + "'")
#         lines = rFile.readlines()
#         rFile.close()
#         
#         self.systemDefFile = fileName
#         self.tileSystemDef = ast.literal_eval(lines[0])
#         self.tileSystemDef['MainFile'] = fileName
#         self.tileSystemDef['DirName'] = os.path.dirname(fileName)
#         
#         print("Read tile system " + str(self.tileSystemDef))
# 
#         midPt = self.LoadAssembly(fileName=fileName)
#         return midPt
#     
#     def SaveSAAssembly(self, fileName, tdsFileName):
#         wFile = open(fileName, "w")
# 
#         wFile.write(tdsFileName + "\n")
#         wFile.write("Temperature=6\n")
# 
#         count = 0
#         for x in range(self.minX, self.maxX+1):
#             for y in range(self.minY, self.maxY+1):
#                 loc = self.GetLoc((x,y))
#                 if not loc is None:
#                     wFile.write(loc.type.name + "-SEED " + str(x) + " " + str(y) + "\n")
#                     #wFile.write(loc.type.name + " " + str(x) + " " + str(y) + "\n")
#                     count = count + 1
#         wFile.close()
#         SimUtils.globalApp.OutputMessage("Saved assembly of " + str(count) + " tiles to file '" + fileName + "'")
#     
#     def AssemblyFromString(self, assemblyString):
#         assemblyString = assemblyString.replace("{", "[")
#         assemblyString = assemblyString.replace("}", "]")
#         assemblyPreList = []
#         if len(assemblyString) > 0:
#             bException = True
#             while bException == True:
#                 try:
#                     assemblyPreList = eval(assemblyString)
#                     bException = False
#                 except NameError as e:
#                     errStr = str(e)
#                     errArr = errStr.split("\'")
#                     name = errArr[1]
#                     assemblyString = assemblyString.replace(name, "\'" + name + "\'")
#                     bException = True
# 
#         assemblyList = []
#         for entry in assemblyPreList:
#             name = entry[0]
#             x = int(entry[1][0][0])
#             y = int(entry[1][0][1])
#             orient = entry[1][1]
#             name = name + "_" + str(orient)
#             assemblyList.append((name, (x,y)))
#         return assemblyList
#     
#     def LoadAssembly(self, seedDefn=None, fileName=None):
#         if not fileName is None:
#             os.stat(fileName)
#             rFile = open(fileName, "r")
#             lines = rFile.readlines()
#             rFile.close()
#             seedDefn = ""
#             for line in lines:
#                 seedDefn = seedDefn + line.strip()
#             
#         self.seedPatch = self.AssemblyFromString(seedDefn)
#         self.ClearSpace()
#         self.frontier = set()
#         (aveX, aveY) = self.PlaceSeed()
#         SimUtils.globalApp.OutputMessage("Loaded assembly with " + str(len(self.seedPatch)) + " tiles")
#         return (aveX, aveY)
# 
#     def AddTile(self, ttype, coords):
#         retVal = Assembly.AddTile(self, ttype, coords)
#         self.tileTypesUsed.add(ttype.name)
#         return retVal
#         
#     def PrintStats(self):
#         SimUtils.globalApp.OutputMessage("Polyomino set size: " + str(len(self.polyomino_types.keys())))
#         SimUtils.globalApp.OutputMessage("polyominoes placed: " + str(len(self.polyomino_list)))
#===============================================================================

