from SimulatorFramework import SimUtils
from SimulatorFramework import Frontier
from SimulatorFramework import SimCanvas
from SimulatorFramework import TileAssembly
from SimulatorFramework import SimApp
from PolyominoTASTileType import *
from PolyominoTASAssembly import *
from PolyominoTASSystem import *
import wx
import wx.adv
#from test.test_support import temp_cwd
import datetime
import xml.etree.ElementTree as ET
import xml.dom.minidom
import random
import copy


IMG_OFFSET_HORIZ = 17
IMG_OFFSET_VERT  = 17

BLOCK_DIMS = 96

class PolyominoTASCanvas(SimCanvas.Canvas):
    def __init__(self, parent, idIn, frame, size, app):
        self.drawBitmaps = True
        self.drawSignals = False
        self.drawUpdates = SimCanvas.SIMFW_FULLDISPLAY
        self.polyominoSystem = PolyominoSystem()
        self.assembly = self.polyominoSystem.assembly
        self.seed = self.polyominoSystem.seed
        SimCanvas.Canvas.__init__(self, parent, idIn, frame, size, app)
        self.imgOffsetHoriz = IMG_OFFSET_HORIZ
        self.imgOffsetVert = IMG_OFFSET_VERT
        self.assembly = PolyominoAssembly()
        self.Bind(wx.EVT_CHAR,self.OnChar)
        
        self.systemDefFile = None

        # Just a counter used for some basic testing
        self.test_counter = 0

    def CreateNewAssembly(self):
        self.sideLen = SimCanvas.CELLSIZE
        self.assembly = PolyominoAssembly()
        self.UpdateSizeVars()
        self.RedrawAssembly()
        self.CanvasSizeUpdate((0,0))
        self.temp_polyominoes = []

    def Reset(self, bFromFiles=True):
        #dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
        dlg = wx.MessageDialog(self, "Are you sure you want to reset your system\nand abandon the current assembly?", "Confirm reset", style=wx.OK|wx.CANCEL|wx.CENTRE|wx.ICON_WARNING)
        if not dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            return
        #result = int(dlg.GetValue())
        dlg.Destroy()
        #self.SetRandomSeed(result)

        self.app.OutputMessage("Reseting system...")

        self.InitializeDrawing()
        
        midPt = (0,0)
        if bFromFiles == True:
            if not self.systemDefFile is None:
                #midPt = self.assembly.LoadTileSystem(self.systemDefFile)
                self.DoLoad(self.systemDefFile)
                return
            else:
                #self.app.OutputMessage("No system definition file to reload")
                SimUtils.globalApp.ShowErrorDialog("No system definition file to reload")
                return
        else:
            self.assembly = copy.deepcopy(self.seed)
            #midPt = self.assembly.PlaceSeed()

        self.RedrawAssembly(checkAll=True)
        self.ScrollToCoords(midPt)

        #self.UpdateSizeVars()
        #self.CanvasSizeUpdate(midPt)
        
        self.app.OutputMessage("Assembly reset to seed")

    def DoSetAsSeed(self):
        self.seed = copy.deepcopy(self.assembly)
        self.app.OutputMessage("Current assembly set as seed")

    def DoLoad(self, fileName=None):
        if fileName is None:
            openFileDialog = wx.FileDialog(self.app.window, "Open", "", "", 
                                           "Polyomino system files (*.xml)|*.xml",
                                           wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
            openFileDialog.ShowModal()
            fileName = openFileDialog.GetPath()
            openFileDialog.Destroy()
        if (len(fileName) > 0):
            
            dlg = wx.TextEntryDialog(self, "You can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
            if dlg.ShowModal() == wx.ID_OK:
                self.randSeed = int(dlg.GetValue())
                dlg.Destroy()
            else:
                dlg.Destroy()
                return

            self.SetRandomSeed(self.randSeed)
            SimUtils.globalApp.OutputMessage("Loading Polyomino system...")
            self.InitializeDrawing()
            self.CreateNewAssembly()
            pcs = PolyominoSystem()
            if pcs.ReadFromFile(fileName) == False:
                self.app.OutputMessage("Failed to load Polyomino system from " + fileName)
                return
            self.systemDefFile = fileName
            self.polyominoSystem = pcs
            self.seed = pcs.seed
            self.assembly = pcs.assembly
            if (self.seed.GetNumTiles() == 0):
                self.seed = copy.deepcopy(self.assembly)
            self.app.OutputMessage(f"Loaded assembly with {self.assembly.GetNumTiles()} polyominoes")
            midPt = (0,0)
            self.RedrawAssembly(checkAll=True)
            self.ScrollToCoords(midPt)
            SimUtils.globalApp.OutputMessage("Polyomino system loaded")

    def DoLoadAssembly(self):
        openFileDialog = wx.FileDialog(self.app.window, "Open", "", "", 
                                       "Assembly files (*.txt)|*.txt",
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            midPt = self.assembly.LoadAssembly(fileName=fileName)
            if midPt == False:
                self.app.OutputMessage("Failed to load assembly from " + fileName)
                return
            self.UpdateSizeVars()
            self.RedrawAssembly(checkAll=True)
            self.CanvasSizeUpdate(midPt)
            self.app.OutputMessage("Loaded assembly from " + fileName)

    def DoSetSeed(self):
        newSeedStr = ""
        msg = "Enter a seed definition"
        dlg = wx.TextEntryDialog(self, msg, caption="Set seed", value=str(newSeedStr))
        if not dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            return

        newSeedStr = dlg.GetValue()
        newSeedStr = str(newSeedStr.strip())
        dlg.Destroy()
        
        midPt = self.assembly.LoadAssembly(seedDefn=newSeedStr)
        if midPt == False:
            self.app.OutputMessage("Failed to set seed assembly to '" + newSeedStr + "'")
            return
        self.RedrawAssembly(checkAll=True)
        self.ScrollToCoords(midPt)
        self.app.OutputMessage("Seed definition set to " + newSeedStr)
        
    def DoSaveAs(self):
        saveFileDialog = wx.FileDialog(self, "Save system as", "", "", 
                                       "Polyomino system files (*.xml)|*.xml",
                                       wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        if not saveFileDialog.ShowModal() == wx.ID_OK:
            SimUtils.globalApp.ShowErrorDialog("You must provide a file name to save the system into.")
            return False
        fileName = saveFileDialog.GetPath()
        saveFileDialog.Destroy()
        if (fileName is None) or len(fileName) < 1:
            return False
        else:
            if self.polyominoSystem.WriteToFile(fileName) == True:
                #self.tileSetFile = fileName
                self.app.OutputMessage("Successfully saved Polyomino system to file " + str(fileName))
            else:
                self.app.OutputMessage("ERROR: Failed to save Polyomino system to file " + str(fileName))

    def PrintMessage(self):
        self.PrintGrowth()
        
    def ShowHelp(self):
        info = wx.adv.AboutDialogInfo()
        info.SetName("Polyomino Tiling Simulator")
        info.SetVersion(self.app.GetVersionString())
        helpStr = "\nList of commands:\n\n"
        helpStr = helpStr + "  l : select tile system definition file to load\n"
        helpStr = helpStr + "  s : save current assembly\n"
        helpStr = helpStr + "  r : reset current tile system to seed\n"
        helpStr = helpStr + "  n : set the seed definition"
        helpStr = helpStr + "  f : simulate one step\n"
        helpStr = helpStr + "  g : begin fast forward simulation\n"
        helpStr = helpStr + "  q : quit fast forward simulation\n"
        helpStr = helpStr + "  p : print list of frontier locations\n"
        helpStr = helpStr + "  t : toggle display of grid lines\n"
        helpStr = helpStr + "  d : toggle display updates of tiles to assembly\n"
        helpStr = helpStr + " +/-: zoom in/out\n"
        helpStr = helpStr + "<left mouse click>:  zoom in on clicked location\n"
        helpStr = helpStr + "<right mouse click>:  zoom out of clicked location\n"
        helpStr = helpStr + "  h : show this help dialog"
        info.SetDescription(helpStr)
        wx.adv.AboutBox(info)

    def OnMouse(self, event):
        x,y = self.ConvertEventCoords(event)
        x2,y2 = self.CoordsFromPixelPos(x, y)

        if self.showMouseOver == True:
            if (x2 >= self.assembly.minX) and (x2 <= self.assembly.maxX) and (y2 >= self.assembly.minY) and (y2 <= self.assembly.maxY):
                self.SetMouseOverText((x2, y2))
            else:
                self.SetMouseOverText(None)
            
        if self.mouseMode == SimCanvas.SIMFW_MODE_SELECT:
            if event.RightUp():
                self.rubberBand.reset()
            else:
                self.rubberBand.HandleMouseEvents(event)
        elif self.mouseMode == SimCanvas.SIMFW_MODE_TILE_ADDREM:
            if event.RightDown():
  
                polyomino = self.assembly.GetLoc((x2, y2))
                if polyomino is None:
                    newPCStr = ""
                    msg = f"Enter a PolyominoType name to be added at location {(x2,y2)}"
                    dlg = wx.TextEntryDialog(self, msg, caption="PolyominoType name", value=str(newPCStr))
                    if not dlg.ShowModal() == wx.ID_OK:
                        dlg.Destroy()
                        return
            
                    newPCStr = dlg.GetValue()
                    newPCStr = str(newPCStr.strip())
                    dlg.Destroy()
                    
                    if len(newPCStr) > 0:
                        if newPCStr in self.polyominoSystem.PolyominoType_dict:
                            new_pc = Polyomino(self.polyominoSystem.PolyominoType_dict[newPCStr])
                            new_pc.translate((x2,y2))
                            self.assembly.AddPolyomino(new_pc)
                            #print(f"Added a Polyomino of type {newPCStr} in location {(x2,y2)}")
                            self.Refresh()
                        else:
                            print(f"ERROR: No PolyominoType with name {newPCStr} exists in the current PolyominoType set.")

                else:
                    #print(f"Gonna delete polyomino {polyomino}")
                    self.assembly.RemovePolyomino(polyomino)
                    self.Refresh()

        elif self.mouseMode == SimCanvas.SIMFW_MODE_DRAW:
            if event.LeftDown() or (event.LeftIsDown() and event.Dragging()):
                loc = self.assembly.GetLoc((x2,y2))
                if loc == None:
                    fLoc = Frontier.FrontierLocation((x2,y2), None)
                    #if fLoc in self.assembly.frontier:
                    self.assembly.disabledFrontier.add(fLoc)
                    self.assembly.SetLoc(fLoc.coords, TileAssembly.FRONTIER_LOC_OFF)
                    self.AddToDrawQueue(([], [], [fLoc], []))
            elif event.RightDown() or (event.RightIsDown() and event.Dragging()):
                fLoc = Frontier.FrontierLocation((x2,y2), None)
                if fLoc in self.assembly.disabledFrontier:
                    self.assembly.disabledFrontier.discard(fLoc)
                    if fLoc in self.assembly.frontier:
                        self.assembly.SetLoc(fLoc.coords, TileAssembly.FRONTIER_LOC)
                        self.AddToDrawQueue(([], [], [fLoc], []))
                    else:
                        self.assembly.ClearLoc(fLoc.coords)
                        self.AddToDrawQueue(([], [], [], [fLoc]))
        else:
            if event.LeftDown():
                if self.mouseMode == SimCanvas.SIMFW_MODE_PAN_ZOOM:
                    self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                    self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                    self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))
    
            elif event.RightDown():
                if self.mouseMode == SimCanvas.SIMFW_MODE_PAN_ZOOM:
                    
                    bZoomIn = True
                    if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                        bZoomIn = False
                    self.DoScale(bZoomIn, bScrollToCenter=False, scrollToCoords=(x2,y2))
                    #self.ScrollToCoords((x2,y2))
                    sz = self.GetSize()
                    self.WarpPointer(sz[0]/2, sz[1]/2)

            elif event.LeftUp():
                self.dragStart = None
                if self.mouseMode == SimCanvas.SIMFW_MODE_PAN_ZOOM:
                    self.SetCursor(wx.Cursor(self.grabCursorImg))
                else:
                    self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            
            elif event.Dragging():
                if self.mouseMode == SimCanvas.SIMFW_MODE_PAN_ZOOM:
                    if self.dragStart is not None:
                        #self.SetScrollRate(1,1)
                        new = self.ScreenToClient(wx.GetMousePosition())
                        dx = int((self.dragStart.x-new.x)/1)
                        dy = int((self.dragStart.y-new.y)/1)
                        spx = self.GetScrollPos(wx.HORIZONTAL)
                        spy = self.GetScrollPos(wx.VERTICAL)
                        self.Scroll(spx+dx, spy+dy)
                        self.dragStart = new

    def SetMouseOverText(self, coords):
        if coords is None:
            self.SendStatusBarUpdate(0, "")
            self.SendStatusBarUpdate(2, "")
            return
        polyominoTypeName = ""
        loc = self.assembly.GetLoc((coords[0], coords[1]))
        if loc is None:
            for tpc in self.temp_polyominoes:
                if tpc[1] == True:
                    c_count = 0
                    for b in tpc[0].blocks:
                        if (coords[0] == b[0]) and (coords[1] == b[1]):
                            loc = (tpc[0], c_count)
                        c_count += 1
        if not loc is None:
            polyomino = loc
            polyominoTypeName = f"Polyomino type: {polyomino.polyominoType.name}"

        else:
            #floc = Frontier.FrontierLocation((coords[0], coords[1]),"")
            if (coords[0], coords[1]) in self.assembly.frontier:
                polyominoTypeName = "Frontier location"
                #print(f'fontier location {self.assembly.frontier[ (coords[0], coords[1])]}')
#                 for f in self.assembly.frontier:
#                     if f == floc:
#                         if f.isGrowthLoc == True:
#                             tileTypeName = "Growth location, allows " + str(f.fittingTypeNames)
#                         else:
#                              
#                             tileTypeName = "Frontier location, allows " + str(next(iter(f.fittingTypeNames)))
            
        msg = "Coords: (" + str(int(coords[0])) + "," + str(int(coords[1])) + ")"
        self.SendStatusBarUpdate(0, msg)
        self.SendStatusBarUpdate(2, polyominoTypeName)

    def DoDrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SimCanvas.SIMFW_NODISPLAY:
            return

        if msg == "all":
            self.Refresh()
            return
        elif msg == "allAndCheck":
        #    self.DoInitializeDrawing(dc)
            self.DoRedrawAssembly()

        self._lock.acquire()
               
        dc = wx.ClientDC(self)
            
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        dc.SetUserScale(self.drawScale, self.drawScale)

        #print("in drawupdate, queue len: " + str(len(self.DrawQueue)))
        if (msg == "new") and self.drawBitmaps == True:
            
            usp = self.CalcUnscrolledPosition(0,0)
            sz = self.GetSize()
            topLeft = self.CoordsFromPixelPos(usp[0], usp[1], True)
            bottomRight = self.CoordsFromPixelPos(usp[0]+sz[0], usp[1]+sz[1], True)

            for e in self.DrawQueue:
                if e[0][0] >= topLeft[0] and e[0][0] <= bottomRight[0] and e[0][1] <= topLeft[1] and e[0][1] >= bottomRight[1]:
                    if e[1] == SimCanvas.SIMFW_DRAWQUEUE_DRAWTILE:
                        loc = self.assembly.GetLoc(e[0])
                        self.DrawAtCoords(dc, e[0][0], e[0][1], loc)
                    elif e[1] == SimCanvas.SIMFW_DRAWQUEUE_DRAWFRONTIER:
                        loc = self.assembly.GetLoc(e[0], includeFrontier=True)
                        self.DrawAtCoords(dc, e[0][0], e[0][1], loc)
                    elif (e[1] == SimCanvas.SIMFW_DRAWQUEUE_REMOVETILE) or (e[1] == SimCanvas.SIMFW_DRAWQUEUE_REMOVEFRONTIER):
                        self.ClearCoord(dc, e[0][0], e[0][1])
                    
            self.DrawQueue = []        
        self._lock.release()
        self.needsRedraw = False

    def OnPaint(self, event):
        self._lock.acquire()

        dc = wx.PaintDC(self)

        bg = wx.Brush("WHITE")#(self.GetBackgroundColour())
        dc.SetBackground(bg)
        dc.Clear()

        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        x, y = self.GetViewStart()
        w, h = self.GetSize()
        topLeft = self.CoordsFromPixelPos(x, y)
        bottomRight = self.CoordsFromPixelPos(x + w, y + h)

        dc.SetUserScale(self.drawScale, self.drawScale)
        
        if self.drawGrid == True:
            self.DrawGrid(dc)

        if self.drawBitmaps == True:
            
            zero_x, zero_y = self.GetPixelsForCoords(0, 0)
            #zero_x = zero_x - int(1 * self.sideLen)
            #zero_y = zero_y - int(1 * self.sideLen)
            
            drawn_polyominoes = []
            for coords in self.assembly.locationMap:
                polyomino = self.assembly.locationMap[coords]
                if not polyomino in drawn_polyominoes:
                    drawn_polyominoes.append(polyomino)

                    #MJP TODO: should make sure it's in the visible portion of the canvas
                    #for i in range(topLeft[0], bottomRight[0]+1):
                        #for j in range(bottomRight[1], topLeft[1]+1):
                    #self.DrawPolyomino(dc, polyomino)
                    polyomino.Draw(dc, (zero_x,zero_y), self.sideLen, self.drawScale)
                
            for (polyomino, bVisible) in self.temp_polyominoes:
                if bVisible == True:
                    polyomino.Draw(dc, (zero_x,zero_y), self.sideLen, self.drawScale)
                
        self._lock.release()
        self.needsRedraw = False
        self.SendStatusBarUpdate(1, "Number of Polyominos: " + str(self.assembly.GetNumTiles()))
        return
    
    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoord = xCoord - self.assembly.minX
        yCoord = yCoord - self.assembly.minY
        xPixel = (xCoord * self.sideLen)# * self.drawScale
        yPixel = (self.height - ((yCoord+1) * self.sideLen))# * self.drawScale
        return (xPixel,yPixel)
  
    def CoordsFromPixelPos(self, x, y, bClosest=True):
        # bClosest isn't used in this default implementation but may be useful in grids where some pixels aren't on valid coordinates
        x1 = int(x / (self.sideLen*self.drawScale))
        y1 = int(((self.height*self.drawScale)-y) / (self.sideLen*self.drawScale))

        x1 = x1 + self.assembly.minX
        y1 = y1 + self.assembly.minY

        return (x1,y1)
    
    def EnableSelectedFrontier(self):
        vis_num = -1
        count = 0
        for (polyomino, bVisible) in self.temp_polyominoes:
            if bVisible:
                vis_num = count
            count += 1
        vis_num = (vis_num + 1) % len(self.temp_polyominoes)
            
        count = 0
        for (polyomino, bVisible) in self.temp_polyominoes:
            if count == vis_num:
                self.temp_polyominoes[count] = (polyomino, True)
            else:
                self.temp_polyominoes[count] = (polyomino, False)
            count += 1
        self.Refresh()
        
    def DisableSelectedFrontier(self):
        for (polyomino, bVisible) in self.temp_polyominoes:
            if bVisible:
                self.assembly.AddPolyomino(polyomino)
        self.temp_polyominoes.clear()
        self.app.toolbar.EnableTool(SimApp.SIMFW_ID_DISABLEFRONTIER, False)
        self.app.toolbar.EnableTool(SimApp.SIMFW_ID_ENABLEFRONTIER, False)

    def DoForwardStep(self):
        savedMinX = self.assembly.minX
        savedMinY = self.assembly.minY
        savedMinZ = self.assembly.minZ
        self._lock.acquire()
        
        new_polyominoes = self.assembly.StepForward()
        
        if not new_polyominoes is None:
            self._lock.release()
            self.DrawAssemblyUpdate("all")
            return True
        else:
            self.app.OutputMessage(f"There are no polyomino types which attach to the current assembly")
            self._lock.release()
            return False

#                 attachments = self.assembly.GetAttachments(pct)
#                 print(attachments)
#                 valid = []
#                 for attachment in attachments:
#                     valid.extend(self.assembly.GetValidPolyominos(pct, attachment))
#     
#                 if len(valid) > 0:
#                     attachment_dict[pct_name] = valid
#                     num_options += len(valid)
# 
# 
#         self.app.OutputMessage(f"There are {num_options} binding options. {len(attachment_dict)} PolyominoTypes can validly attach to the current assembly")
#             
#         rando = -1
#         if self.keepGoing and (num_options > 1):  # we're in fast forward mode, randomly select an option
#             rando = random.randrange(num_options)
#             
#         count = 0
#         bVis = True
#         for pct_name in attachment_dict:
#             attachments = attachment_dict[pct_name]
#             self.app.OutputMessage(f"PolyominoType '{pct_name}' can bind in {len(attachments)} configurations:")
#             for attachment in attachments:
#                 self.app.OutputMessage(f"\t{attachment}")
#                 if (num_options == 1) or (count == rando):
#                     # Automatically add if there's only one option or it was the randomly selected one cuz we're in fast forward
#                     num_options = 1
#                     self.assembly.AddPolyomino(attachment)
#                 else:
#                     self.temp_polyominoes.append((attachment, bVis))
#                     bVis = False
#                 count += 1
#         
#         if num_options > 1:
#             self.app.toolbar.EnableTool(SimApp.SIMFW_ID_DISABLEFRONTIER, True)
#             self.app.toolbar.EnableTool(SimApp.SIMFW_ID_ENABLEFRONTIER, True)
#         else:
#             self.app.toolbar.EnableTool(SimApp.SIMFW_ID_DISABLEFRONTIER, False)
#             self.app.toolbar.EnableTool(SimApp.SIMFW_ID_ENABLEFRONTIER, False)
# 
#         self._lock.release()
#         self.DrawAssemblyUpdate("all")
#         retVal = (num_options > 0)
# 
#         return retVal
#     
#         if not assemblyUpdates is None:
#             if not self.width == (self.assembly.width * self.sideLen) or not self.height == (self.assembly.height * self.sideLen):
#                 centerCoords = self.GetCenterCoords()
#                 coords = (centerCoords[0] + savedMinX - self.assembly.minX, centerCoords[1] + savedMinY - + self.assembly.minY)
#                 #TODO: There is a bug related to recentering after canvas growth. Perhaps it is related to additional growth of the assembly before canvas resizing
#                 print(f'About to update canvas size, centerCoords = {centerCoords}, savedMins = {(savedMinX,savedMinY)}, assembly mins = {(self.assembly.minX,self.assembly.minY)}, computed coords = {coords}')
#                 self.UpdateSizeVars()
#                 self.CanvasSizeUpdate(coords)
#             else:
#                 self.AddToDrawQueue(assemblyUpdates[0])
#             self.CheckForBreakpointHit(assemblyUpdates[0][0][0])
#             if self.drawUpdates == SimCanvas.SIMFW_NODISPLAY:
#                 if (self.assembly.numTiles % 1000) == 0:
#                     
#                     self.SendStatusBarUpdate(1, "Number of Polyominos: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
#             elif (self.keepGoing == False) or (self.assembly.numTiles % 1000) == 0:
#                 #MJP TODO: handle this better, but it REALLY slows things down!!!
#                 self.SendStatusBarUpdate(1, "Number of Polyominos: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
#             return assemblyUpdates[1]
#         return False

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("n") or keycode == ord("N"):
            self.DoSetSeed()
        elif keycode == ord("a") or keycode == ord("A"):
            self.DoLoadAssembly()
        elif keycode == ord("z") or keycode == ord("Z"):
            self.assembly.PrintStats()
        elif keycode == ord("x") or keycode == ord("X"):
            self.TestPolyominos()
        else:
            self.app.OnChar(evt)
            
    def TestPolyominos(self):

        self.test_counter += 1# (self.test_counter + 1) % 3
        self.ScrollToCoords((0,0))
        self.Refresh()
